import json as jsonlib
from http.cookiejar import CookieJar

from geventhttpclient import useragent


class RequestsRequest(useragent.CompatRequest):
    @property
    def body(self):
        return self.payload


class RequestsResponse(useragent.CompatResponse):
    @property
    def request(self):
        return self._request

    @property
    def ok(self):
        return 100 <= self.status_code < 400

    @property
    def reason(self):
        return self._response.status_message

    @property
    def url(self):
        return self._request.url

    @property
    def is_redirect(self):
        """True if this Response is a well-formed HTTP redirect that could have
        been processed automatically (by :meth:`Session.resolve_redirects`).
        """
        return "location" in self.headers and self.status_code in range(300, 310)

    @property
    def raw(self):
        return self.stream

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            raise useragent.BadStatusCode(self.url, code=self.status_code)


class Session(useragent.UserAgent):
    """This class mimics and blatantly borrows with all due respect
    from the excellent and rightfully popular Requests API.

    Copyright of the original requests project:

    :copyright: (c) 2012 by Kenneth Reitz.
    :license: Apache2, see LICENSE for more details.
    """

    request_type = RequestsRequest
    response_type = RequestsResponse

    def get(self, url, **kw):
        r"""Sends a GET request. Returns a HTTP Response object.

        :param url: URL for the new a HTTP Request object.
        :param \*\*kw: Optional arguments that ``request`` takes.
        :rtype: CompatResponse
        """

        kw.setdefault("allow_redirects", True)
        return self.request("GET", url, **kw)

    def options(self, url, **kw):
        r"""Sends a OPTIONS request. Returns a HTTP Response object.

        :param url: URL for the new a HTTP Request object.
        :param \*\*kw: Optional arguments that ``request`` takes.
        :rtype: CompatResponse
        """

        kw.setdefault("allow_redirects", True)
        return self.request("OPTIONS", url, **kw)

    def head(self, url, **kw):
        r"""Sends a HEAD request. Returns a HTTP Response object.

        :param url: URL for the new a HTTP Request object.
        :param \*\*kw: Optional arguments that ``request`` takes.
        :rtype: CompatResponse
        """

        kw.setdefault("allow_redirects", False)
        return self.request("HEAD", url, **kw)

    def post(self, url, data=None, json=None, **kw):
        r"""Sends a POST request. Returns a HTTP Response object.

        :param url: URL for the new a HTTP Request object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the HTTP Request.
        :param json: (optional) json to send in the body of the HTTP Request.
        :param \*\*kw: Optional arguments that ``request`` takes.
        :rtype: CompatResponse
        """

        return self.request("POST", url, data=data, json=json, **kw)

    def put(self, url, data=None, **kw):
        r"""Sends a PUT request. Returns a HTTP Response object.

        :param url: URL for the new a HTTP Request object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the HTTP Request.
        :param \*\*kw: Optional arguments that ``request`` takes.
        :rtype: CompatResponse
        """

        return self.request("PUT", url, data=data, **kw)

    def patch(self, url, data=None, **kw):
        r"""Sends a PATCH request. Returns a HTTP Response object.

        :param url: URL for the new a HTTP Request object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the HTTP Request.
        :param \*\*kw: Optional arguments that ``request`` takes.
        :rtype: CompatResponse
        """

        return self.request("PATCH", url, data=data, **kw)

    def delete(self, url, **kw):
        r"""Sends a DELETE request. Returns a HTTP Response object.

        :param url: URL for the new a HTTP Request object.
        :param \*\*kw: Optional arguments that ``request`` takes.
        :rtype: CompatResponse
        """

        return self.request("DELETE", url, **kw)

    def request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
    ):
        """Constructs and sends a HTTP request, returns a HTTP Response object.

        NOTE: Only a subset of these parameters is currently (fully) supported.
        And it's also not

        :param method: method for the new request object.
        :param url: URL for the new request object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the request.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the request.
        :param json: (optional) json to send in the body of the
            HTTP Request.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            HTTP Request.
        :param cookies: (optional) Dict or CookieJar object to send with the
            HTTP Request.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :type timeout: float or tuple
        :param allow_redirects: (optional) Set to True by default.
        :type allow_redirects: bool
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair.
        :rtype: CompatResponse
        """
        for param_name, param in dict(timeout=timeout, cert=cert, verify=verify).items():
            if param is not None:
                raise ValueError(
                    f"{param} can not be set on a per-request basis. Please configure the UserAgent instead."
                )
        for param_name, param in dict(
            cookies=cookies, auth=auth, proxies=proxies, hooks=hooks
        ).items():
            if param is not None:
                raise NotImplementedError(
                    f"{param} is currently unsupported as a keyword argument."
                )
        for param_name, param in dict(hooks=hooks).items():
            if param is not None:
                raise NotImplementedError(f"{param} is not supported")

        if json:
            if data:
                raise ValueError("Can send either data or json, not both at once")
            data = jsonlib.dumps(json)
            if headers is None:
                headers = {}
            headers["Content-Type"] = "application/json"

        response = self.urlopen(
            url,
            method=method.upper(),
            headers=headers,
            files=files,
            payload=data or None,
            params=params,
            max_redirects=None if allow_redirects else 0,
        )
        if stream is False:
            # preload the data
            _ = response.content
        return response

    def __init__(self, *args, **kw):
        """
        requests.Session has no arguments at all. Unfortunately, we're relying way more
        on configuring the Session / UserAgent, while requests focuses more on
        configuring the single requests.
        """
        kw.setdefault("max_redirects", 30)
        super().__init__(*args, **kw)
        if not self.cookiejar:
            self.cookiejar = CookieJar()

    def _verify_status(self, status_code, url=None):
        # Don't raise, whatever the status is
        pass



"""
Drop-in replacement for httplib2. Make sure to monkey patch everything first!

import gevent.monkey
gevent.monkey.patch_all()

from geventhttpclient import httplib
httplib.patch()

from geventhttpclient import httplib2

http = httplib2.Http(concurrency=5)
"""

from contextlib import contextmanager

import gevent.queue

import geventhttpclient.httplib


class ClientPool:
    """
    Pool implementation for HTTP clients, that weren't designed with concurrency in mind.
    Usage example:

    pool = ClientPool(MyHttpClient)
    with pool.get() as client:
        response = client.request("127.0.0.1")
    """

    def __init__(self, factory, concurrency=5):
        self.factory = factory
        self.queue = gevent.queue.Queue(concurrency)
        for i in range(concurrency):
            self.queue.put(factory())

    @contextmanager
    def get(self):
        client = self.queue.get()
        yield client
        self.queue.put(client)

    def close(self):
        for client in self.queue:
            client.close()


with geventhttpclient.httplib.patched():
    import httplib2

    class Http:
        """Imitate a httplib2.Http client. Except that it is now run concurrently."""

        def __init__(self, *args, **kw):
            self.concurrency = max(kw.pop("concurrency", 0), 2)  # never smaller than 2
            self.args = args
            self.kw = kw
            self.pool = ClientPool(self._http_factory, concurrency=self.concurrency)

        def _http_factory(self):
            return httplib2.Http(*self.args, **self.kw)

        def request(self, *args, **kw):
            with self.pool.get() as client:
                return client.request(*args, **kw)

        def close(self):
            self.pool.close()



import errno
import json as jsonlib
import os
import socket
import ssl
import sys
import zlib
from urllib.parse import urlencode

import brotli
import gevent
from urllib3 import encode_multipart_formdata
from urllib3.fields import RequestField

from geventhttpclient.client import HTTPClient, HTTPClientPool
from geventhttpclient.url import URL, to_key_val_list


class ConnectionError(Exception):
    def __init__(self, url, *args, **kw):
        self.url = url
        self.__dict__.update(kw)
        if args and isinstance(args[0], str):
            self.text = args[0] + ": " + str(args[1:])
        else:
            self.text = str(args[0]) if len(args) == 1 else ""
        if kw:
            self.text += ", " if self.text else ""
            self.kw_text = ", ".join(f"{key}={val}" for key, val in kw.items())
            self.text += self.kw_text
        else:
            self.text = ""

    def __str__(self):
        if self.text:
            return f"URL {self.url}: {self.text}"
        else:
            return f"URL {self.url}"

    def __repr__(self):
        repr_str = super().__repr__()
        if self.kw_text:
            return repr_str.replace(")", "".join([", ", self.kw_text, ")"]))
        return repr_str


class RetriesExceeded(ConnectionError):
    pass


class BadStatusCode(ConnectionError):
    pass


class EmptyResponse(ConnectionError):
    pass


class CompatRequest:
    """urllib.request.Request compatible request class.
    See also: http://docs.python.org/library/cookielib.html
    """

    # TODO: Since we dropped all py2 dependencies, we could derive directly from
    #       urllib.request.Request and only deviate when required. Not rebuild
    #       the whole thing.

    def __init__(self, url, method="GET", headers=None, payload=None, params=None):
        self.set_url(url, params=params)
        self.original_host = self.url_split.host
        self.method = method.upper()
        self.headers = headers
        self.payload = payload

    @property
    def full_url(self):
        # new in python3.x
        return self.url

    def set_url(self, url, params=None):
        if isinstance(url, URL):
            self.url = str(url)
            self.url_split = url
        else:
            self.url = url
            self.url_split = URL(self.url, params=params)

    def get_full_url(self):
        return self.url

    def get_host(self):
        return self.url_split.host

    def get_type(self):
        return self.url_split.scheme

    def get_origin_req_host(self):
        return self.original_host

    def is_unverifiable(self):
        """See http://tools.ietf.org/html/rfc2965.html. Not fully implemented!"""
        return False

    @property
    def unverifiable(self):
        return self.is_unverifiable()

    def get_header(self, header_name, default=None):
        return self.headers.get(header_name, default)

    def has_header(self, header_name):
        return header_name in self.headers

    def header_items(self):
        return list(self.headers.items())

    def add_unredirected_header(self, key, val):
        self.headers.add(key, val)

    def _drop_payload(self):
        self.method = "GET"
        self.payload = None
        for item in ("content-length", "content-type", "content-encoding"):
            self.headers.discard(item)

    def _drop_cookies(self):
        for item in ("cookie", "cookie2"):
            self.headers.discard(item)

    def redirect(self, code, location):
        """Modify the request inplace to point to the new location"""
        self.set_url(self.url_split.redirect(location))
        if code in (302, 303):
            self._drop_payload()
        self._drop_cookies()


class CompatResponse:
    """Adapter for urllib3-style responses."""

    __slots__ = "headers", "_response", "_request", "_sent_request", "_cached_content"

    def __init__(self, ghc_response, request=None, sent_request=None):
        self._response = ghc_response
        self._request = request
        self._sent_request = sent_request
        self.headers = self._response._headers_index

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.release()

    @property
    def status_code(self):
        """HTTP status code as plain integer"""
        return self._response.get_code()

    def __len__(self):
        """The content lengths as declared from the headers"""
        return self._response.length

    def info(self):
        """Adaption to http.client."""
        return self.headers

    def __nonzero__(self):
        """If we have an empty response body, we still don't want to evaluate as false"""
        return True

    def __iter__(self):
        return iter(self._response)

    def read(self, n=None):
        """Read n bytes from the response body"""
        return self._response.read(n)

    def readline(self):
        return self._response.readline()

    def release(self):
        return self._response.release()

    def unzipped(self, gzip=True, br=False):
        bodystr = self._response.read()
        if gzip:
            return zlib.decompress(bodystr, 16 + zlib.MAX_WBITS)
        elif br:
            return brotli.decompress(bodystr)
        else:
            # zlib only provides the zlib compress format, not the deflate format;
            # so on top of all there's this workaround:
            try:
                return zlib.decompress(bodystr, -zlib.MAX_WBITS)
            except zlib.error:
                return zlib.decompress(bodystr)

    @property
    def content(self):
        """Unzips if necessary and buffers the received body. Careful with large files!"""
        try:
            return self._cached_content
        except AttributeError:
            self._cached_content = self._content()
            return self._cached_content

    def _content(self):
        try:
            content_encoding = self.headers.getheaders("content-encoding")[0].lower()
        except IndexError:
            # No content-encoding header set
            content_encoding = "identity"

        if content_encoding == "gzip":
            ret = self.unzipped(gzip=True)
        elif content_encoding == "deflate":
            ret = self.unzipped(gzip=False)
        elif content_encoding == "identity":
            ret = self._response.read()
        elif content_encoding == "br":
            ret = self.unzipped(gzip=False, br=True)
        elif content_encoding == "compress":
            raise ValueError(f"Compression type not supported: {content_encoding}")
        else:
            raise ValueError(f"Unknown content encoding: {content_encoding}")

        self.release()
        return ret

    @property
    def text(self):
        if not self.content:
            return ""

        try:
            content_type = self.headers.getheaders("content-type")[0].lower()
        except IndexError:
            # No content-encoding header set, let's hope for the best
            return self.content.decode()

        if content_type.startswith("text"):
            codec = "utf-8"  # default
            if "charset" in content_type:
                try:
                    codec = content_type.split("charset=", 1)[1][:10]
                except IndexError:
                    pass
            return self.content.decode(codec)
        return self.content

    def json(self):
        return jsonlib.load(self)

    # the stuff only for urllib3

    @property
    def status(self):
        """HTTP status for urllib3"""
        return str(self.status_code)

    @property
    def data(self):
        """Content for urllib3"""
        return self.content

    @property
    def stream(self):
        """Readable stream for urllib3"""
        return self._response

    def isclosed(self):
        """Closed status for urllib3"""
        return self._response.message_complete


class UserAgent:
    response_type = CompatResponse
    request_type = CompatRequest
    valid_response_codes = frozenset([200, 206, 301, 302, 303, 307])
    redirect_resonse_codes = frozenset([301, 302, 303, 307])

    def __init__(
        self,
        max_redirects=3,
        max_retries=3,
        retry_delay=0,
        cookiejar=None,
        headers=None,
        **kw,
    ):
        self.max_redirects = int(max_redirects)
        self.max_retries = int(max_retries)
        self.retry_delay = retry_delay
        self.default_headers = HTTPClient.DEFAULT_HEADERS.copy()
        if headers:
            self.default_headers.update(headers)
        self.cookiejar = cookiejar
        self.clientpool = HTTPClientPool(**kw)

    def close(self):
        self.clientpool.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _verify_status(self, status_code, url=None):
        """Hook for subclassing"""
        if status_code not in self.valid_response_codes:
            raise BadStatusCode(url, code=status_code)

    def _handle_error(self, e, url=None):
        """Hook for subclassing. Raise the error to interrupt further retrying,
        return it to continue retries and save the error, when retries
        exceed the limit.
        Temporary errors should be swallowed here for automatic retries.
        """
        if isinstance(e, (socket.timeout, gevent.Timeout)):
            return e
        elif isinstance(e, socket.error) and e.errno in {
            errno.ETIMEDOUT,
            errno.ENOLINK,
            errno.ENOENT,
            errno.EPIPE,
        }:
            return e
        elif isinstance(e, ssl.SSLError) and "read operation timed out" in str(e):
            return e
        elif isinstance(e, EmptyResponse):
            return e
        raise e.with_traceback(sys.exc_info()[2])

    def _handle_retries_exceeded(self, url, last_error=None):
        """Hook for subclassing"""
        raise RetriesExceeded(url, self.max_retries, original=last_error)

    def urlopen(
        self,
        url,
        method="GET",
        response_codes=valid_response_codes,
        headers=None,
        payload=None,
        to_string=False,
        debug_stream=None,
        params=None,
        max_retries=None,
        max_redirects=None,
        files=None,
        **kw,
    ):
        """Open a URL, do retries and redirects and verify the status code"""
        # POST or GET parameters can be passed in **kw
        req_headers = self.default_headers.copy()
        if headers:
            req_headers.update(headers)
        if kw:
            if method.upper() == "POST":
                if payload is None:
                    payload = kw
                elif isinstance(payload, dict):
                    payload.update(kw)
            else:
                if params is None:
                    params = kw
                elif isinstance(params, dict):
                    params.update(kw)
        req = _make_request(
            url,
            method,
            headers=req_headers,
            payload=payload,
            params=params,
            files=files,
            request_type=self.request_type,
        )
        max_retries = int(max_retries) if max_retries is not None else self.max_retries
        max_redirects = int(max_redirects) if max_redirects is not None else self.max_redirects

        for retry in range(max_retries + 1):
            if retry > 0 and self.retry_delay:
                # Don't wait the first time and skip if no delay specified
                gevent.sleep(self.retry_delay)
            for _ in range(max_redirects + 1):
                if self.cookiejar is not None:
                    self.cookiejar.add_cookie_header(req)

                try:
                    resp = self._urlopen(req)
                except gevent.GreenletExit:
                    raise
                except BaseException as e:
                    e.request = req
                    last_error = self._handle_error(e, url=req.url)
                    break  # Continue with next retry

                # We received a response
                if debug_stream is not None:
                    debug_stream.write(
                        self._conversation_str(req.url, resp, payload=req.payload) + "\n\n"
                    )

                if self.cookiejar is not None:
                    self.cookiejar.extract_cookies(resp, req)

                try:
                    self._verify_status(resp.status_code, url=req.url)
                except Exception as e:
                    # Basic transmission successful, but not the wished result
                    # Let's collect some debug info
                    e.response = resp
                    e.request = req
                    e.http_log = self._conversation_str(req.url, resp, payload=req.payload)
                    resp.release()
                    last_error = self._handle_error(e, url=req.url)
                    break  # Continue with next retry

                redirection = resp.headers.get("location")
                if isinstance(redirection, bytes):
                    redirection = redirection.decode("utf-8")
                if resp.status_code in self.redirect_resonse_codes and redirection:
                    resp.release()
                    try:
                        req.redirect(resp.status_code, redirection)
                        continue
                    except Exception as e:
                        last_error = self._handle_error(e, url=req.url)
                        break

                if not to_string:
                    return resp
                else:
                    # to_string added as parameter, to handle empty response
                    # bodies as error and continue retries automatically
                    try:
                        ret = resp.content
                    except Exception as e:
                        last_error = self._handle_error(e, url=req.url)
                        break
                    else:
                        if not ret:
                            e = EmptyResponse(url, "Empty response body received")
                            last_error = self._handle_error(e, url=req.url)
                            break
                        else:
                            return ret
            else:
                e = RetriesExceeded(url, f"Redirection limit reached ({self.max_redirects})")
                last_error = self._handle_error(e, url=url)
        else:
            return self._handle_retries_exceeded(url, last_error=last_error)

    def _urlopen(self, request):
        client = self.clientpool.get_client(request.url_split)
        resp = client.request(
            request.method,
            request.url_split.quoted_uri,
            body=request.payload,
            headers=request.headers,
        )
        return self.response_type(resp, request=request, sent_request=resp._sent_request)

    @classmethod
    def _conversation_str(cls, url, resp, payload=None, encoding="utf-8"):
        header_str = "\n".join(f"{key}: {val}" for key, val in resp.headers.items())
        ret = "REQUEST: " + url + "\n" + resp._sent_request
        if payload:
            if isinstance(payload, bytes):
                try:
                    ret += payload.decode(encoding) + "\n\n"
                except UnicodeDecodeError:
                    ret += "UnicodeDecodeError" + "\n\n"
            elif isinstance(payload, str):
                ret += payload + "\n\n"
        ret += (
            "RESPONSE: "
            + resp._response.version
            + " "
            + str(resp.status_code)
            + "\n"
            + header_str
            + "\n\n"
            + resp.content.decode(encoding)
        )
        return ret

    def download(self, url, fpath, chunk_size=16 * 1024, resume=False, **kw):
        kw.pop("to_string", None)
        headers = kw.pop("headers", {})
        headers["Connection"] = "Keep-Alive"
        if resume and os.path.isfile(fpath):
            offset = os.path.getsize(fpath)
        else:
            offset = 0

        for _ in range(self.max_retries + 1):
            if offset:
                headers["Range"] = f"bytes={offset}-"
                resp = self.urlopen(url, headers=headers, **kw)
                cr = resp.headers.get("Content-Range")
                if (
                    resp.status_code != 206
                    or not cr
                    or not cr.startswith("bytes")
                    or not cr.split(None, 1)[1].startswith(str(offset))
                ):
                    resp.release()
                    offset = 0
            if not offset:
                headers.pop("Range", None)
                resp = self.urlopen(url, headers=headers, **kw)

            with open(fpath, "ab" if offset else "wb") as f:
                if offset:
                    f.seek(offset, os.SEEK_SET)
                try:
                    data = resp.read(chunk_size)
                    with resp:
                        while data:
                            f.write(data)
                            data = resp.read(chunk_size)
                except BaseException as e:
                    self._handle_error(e, url=url)
                    if resp.headers.get("accept-ranges") == "bytes":
                        # Only if this header is set, we can fall back to partial download
                        offset = f.tell()
                    continue
            # All done, break outer loop
            break
        else:
            self._handle_retries_exceeded(url, last_error=e)
        return resp

    def _make_request(self, url, method="GET", headers=None, payload=None, params=None, files=None):
        """Backwards compatibility for locust."""
        return _make_request(
            url,
            method=method,
            headers=headers,
            payload=payload,
            params=params,
            files=files,
            request_type=self.request_type,
        )


def _make_request(
    url, method, headers, payload=None, params=None, files=None, request_type=CompatRequest
):
    # Adjust headers depending on payload content
    content_type = headers.get("content-type", None)
    if files:
        payload, content_type = _encode_multipart_formdata(files, payload)
        headers["content-type"] = content_type
        headers["content-length"] = len(payload)
    elif payload:
        if isinstance(payload, dict):
            if not content_type:
                headers["content-type"] = "application/x-www-form-urlencoded; charset=utf-8"
            payload = urlencode(payload).encode()
            headers["content-length"] = len(payload)
        elif not content_type and isinstance(payload, str):
            headers["content-type"] = "text/plain; charset=utf-8"
            payload = payload.encode()
            headers["content-length"] = len(payload)
        elif not content_type:
            headers["content-type"] = "application/octet-stream"

    return request_type(url, method=method, headers=headers, payload=payload, params=params)


def _guess_filename(file):
    """Tries to guess the filename of the given object."""
    name = getattr(file, "name", None)
    if not name or not isinstance(name, (str, bytes)):
        return
    if isinstance(name, bytes):
        name = name.decode()
    if name[0] != "<" and name[-1] != ">":
        return os.path.basename(name)


def _encode_multipart_formdata(files, data):
    """
    Method taken from models in requests library , usage is the same. The only difference is that
    you can add custom boundary in 5-tuple version.

    Build the body for a multipart/form-data request.

    Will successfully encode files when passed as a dict or a list of
    tuples. Order is retained if data is a list of tuples but arbitrary
    if parameters are supplied as a dict.

    The tuples may be
    2-tuples (filename, fileobj),
    3-tuples (filename, fileobj, contentype),
    4-tuples (filename, fileobj, contentype, custom_headers) or
    5-tuples (filename, fileobj, contentype, custom_headers, custom boundary).

    example:
    files = {'file': ('report.xls', body, 'application/vnd.ms-excel', {'Expires': '0'}, 'custom_boundary')}

    """

    if not files:
        raise ValueError("Files must be provided.")
    elif isinstance(data, (str, bytes)):
        raise ValueError("Data must not be a string.")

    new_fields = []
    fields = to_key_val_list(data or {})
    files = to_key_val_list(files or {})

    for field, val in fields:
        if isinstance(val, (str, bytes)) or not hasattr(val, "__iter__"):
            val = [val]
        for v in val:
            if v is not None:
                if not isinstance(v, bytes):
                    v = str(v)

                new_fields.append(
                    (
                        field.decode("utf-8") if isinstance(field, bytes) else field,
                        v.encode("utf-8") if isinstance(v, str) else v,
                    )
                )

    for k, v in files:
        # support for explicit filename
        ft = None
        fh = None
        boundary = None
        if isinstance(v, (tuple, list)):
            if len(v) == 2:
                fn, fp = v
            elif len(v) == 3:
                fn, fp, ft = v
            elif len(v) == 4:
                fn, fp, ft, fh = v
            else:
                fn, fp, ft, fh, boundary = v
        else:
            fn = _guess_filename(v) or k
            fp = v

        if isinstance(fp, (str, bytes, bytearray)):
            fdata = fp
        elif hasattr(fp, "read"):
            fdata = fp.read()
        elif fp is None:
            continue
        else:
            fdata = fp

        rf = RequestField(name=k, data=fdata, filename=fn, headers=fh)
        rf.make_multipart(content_type=ft)
        new_fields.append(rf)

    body, content_type = encode_multipart_formdata(new_fields, boundary)

    return body, content_type



import os

import gevent.queue
import gevent.socket
from gevent import lock

_CA_CERTS = None

try:
    from ssl import get_default_verify_paths
except ImportError:
    _CA_CERTS = None
else:
    _certs = get_default_verify_paths()
    _CA_CERTS = _certs.cafile or _certs.capath

if not _CA_CERTS or os.path.isdir(_CA_CERTS):
    import certifi

    _CA_CERTS = certifi.where()

_DEFAULT_CIPHERS = (
    "ECDH+AESGCM:DH+AESGCM:ECDH+AES256:DH+AES256:ECDH+AES128:DH+AES:ECDH+HIGH:"
    "DH+HIGH:ECDH+3DES:DH+3DES:RSA+AESGCM:RSA+AES:RSA+HIGH:RSA+3DES:ECDH+RC4:"
    "DH+RC4:RSA+RC4:!aNULL:!eNULL:!MD5"  # codespell-ignore
)


DEFAULT_CONNECTION_TIMEOUT = 5.0
DEFAULT_NETWORK_TIMEOUT = 5.0

IGNORED = object()


class ConnectionPool:
    DEFAULT_CONNECTION_TIMEOUT = 5.0
    DEFAULT_NETWORK_TIMEOUT = 5.0

    def __init__(
        self,
        connection_host,
        connection_port,
        request_host,
        request_port,
        size=5,
        disable_ipv6=False,
        connection_timeout=DEFAULT_CONNECTION_TIMEOUT,
        network_timeout=DEFAULT_NETWORK_TIMEOUT,
        use_proxy=False,
    ):
        self._closed = False
        self._connection_host = connection_host
        self._connection_port = connection_port
        self._request_host = request_host
        self._request_port = request_port
        self._semaphore = lock.BoundedSemaphore(size)
        self._socket_queue = gevent.queue.LifoQueue(size)
        self._use_proxy = use_proxy

        self.connection_timeout = connection_timeout
        self.network_timeout = network_timeout
        self.size = size
        self.disable_ipv6 = disable_ipv6

    def _resolve(self):
        """resolve (dns) socket information needed to connect it."""
        family = 0
        if self.disable_ipv6:
            family = gevent.socket.AF_INET
        info = gevent.socket.getaddrinfo(
            self._connection_host,
            self._connection_port,
            family,
            0,
            gevent.socket.SOL_TCP,
        )
        # family, socktype, proto, canonname, sockaddr = info[0]
        return info

    def close(self):
        self._closed = True
        while not self._socket_queue.empty():
            try:
                sock = self._socket_queue.get(block=False)
                try:
                    sock.close()
                except:  # noqa
                    pass
            except gevent.queue.Empty:
                pass

    def _create_tcp_socket(self, family, socktype, protocol):
        """tcp socket factory."""
        sock = gevent.socket.socket(family, socktype, protocol)
        return sock

    def _create_socket(self):
        """might be overridden and super for wrapping into a ssl socket
        or set tcp/socket options
        """
        sock_infos = self._resolve()
        first_error = None
        for sock_info in sock_infos:
            try:
                sock = self._create_tcp_socket(*sock_info[:3])
            except Exception as e:
                if not first_error:
                    first_error = e
                continue

            try:
                sock.settimeout(self.connection_timeout)
                sock = self._connect_socket(sock, sock_info[-1])
                self.after_connect(sock)
                sock.settimeout(self.network_timeout)
                return sock
            except OSError as e:
                sock.close()
                if not first_error:
                    first_error = e
            except:  # noqa
                sock.close()
                raise

        if first_error:
            raise first_error
        else:
            raise RuntimeError(f"Cannot resolve {self._host}:{self._port}")

    def after_connect(self, sock):
        pass

    def _connect_socket(self, sock, address):
        sock.connect(address)
        self._setup_proxy(sock)
        return sock

    def _setup_proxy(self, sock):
        if self._use_proxy:
            sock.send(
                bytes(
                    f"CONNECT {self._request_host}:{self._request_port} HTTP/1.1\r\n\r\n",
                    "utf8",
                )
            )

            resp = sock.recv(4096)
            parts = resp.split()
            if not parts or parts[1] != b"200":
                raise RuntimeError(f"Error response from Proxy server : {resp}")

    def get_socket(self):
        """get a socket from the pool. This blocks until one is available."""
        self._semaphore.acquire()
        if self._closed:
            raise RuntimeError("connection pool closed")
        try:
            return self._socket_queue.get(block=False)
        except gevent.queue.Empty:
            try:
                return self._create_socket()
            except:  # noqa
                self._semaphore.release()
                raise

    def return_socket(self, sock):
        """return a socket to the pool."""
        if self._closed:
            try:
                sock.close()
            except:  # noqa
                pass
            return
        self._socket_queue.put(sock)
        self._semaphore.release()

    def release_socket(self, sock):
        """call when the socket is no more usable."""
        try:
            sock.close()
        except:  # noqa
            pass
        if not self._closed:
            self._semaphore.release()


try:
    from ssl import PROTOCOL_TLS_CLIENT

    import gevent.ssl

    try:
        from gevent.ssl import create_default_context
    except ImportError:
        create_default_context = None

except ImportError:
    pass
else:

    def init_ssl_context(
        ssl_context_factory, ca_certs, check_hostname=True, ssl_options=None
    ) -> gevent.ssl.SSLContext:
        """
        Initializes an SSL context with additional SSL options.

        :param ssl_context_factory: Callable to create an SSL context
        :param ca_certs: Path to CA certificates file
        :param check_hostname: Whether to enable hostname checking
        :param ssl_options: Optional dictionary of additional SSL options
        :return: Configured SSLContext instance
        """
        ssl_options = ssl_options or {}

        try:
            ssl_context = ssl_context_factory(cafile=ca_certs)
        except TypeError:
            ssl_context = ssl_context_factory()
            ssl_context.load_verify_locations(cafile=ca_certs)

        ssl_context.check_hostname = check_hostname
        if check_hostname:
            ssl_context.verify_mode = gevent.ssl.CERT_REQUIRED

        if "certfile" in ssl_options and "keyfile" in ssl_options:
            ssl_context.load_cert_chain(
                certfile=ssl_options["certfile"], keyfile=ssl_options["keyfile"]
            )

        if "ciphers" in ssl_options:
            ssl_context.set_ciphers(ssl_options["ciphers"])

        # Apply additional SSL options (e.g., options, verify_flags)
        for option in ["options", "verify_flags"]:
            if option in ssl_options:
                setattr(ssl_context, option, ssl_options[option])

        return ssl_context

    class SSLConnectionPool(ConnectionPool):
        """SSLConnectionPool creates connections wrapped with SSL/TLS.

        :param host: hostname
        :param port: port
        :param ssl_options: accepts any options supported by `ssl.wrap_socket`
        :param ssl_context_factory: use `ssl.create_default_context` by default
            if provided. It must be a callable that returns a SSLContext.
        """

        default_options = {
            "ciphers": _DEFAULT_CIPHERS,
            "ca_certs": _CA_CERTS,
            "cert_reqs": gevent.ssl.CERT_REQUIRED,
            "ssl_version": PROTOCOL_TLS_CLIENT,
        }

        def __init__(
            self,
            connection_host,
            connection_port,
            request_host,
            request_port,
            insecure=False,
            ssl_context_factory=None,
            ssl_options=None,
            **kw,
        ):
            self.insecure = insecure

            self.ssl_options = self.default_options.copy()
            self.ssl_options.update(ssl_options or {})

            ssl_context_factory = ssl_context_factory or create_default_context
            if ssl_context_factory is not None:
                self.ssl_context = init_ssl_context(
                    ssl_context_factory,
                    self.ssl_options["ca_certs"],
                    check_hostname=not self.insecure,
                    ssl_options=ssl_options,
                )
            else:
                self.ssl_context = None

            super().__init__(connection_host, connection_port, request_host, request_port, **kw)

        def _connect_socket(self, sock, address):
            sock = super()._connect_socket(sock, address)

            if self.ssl_context is None:
                # create_default_context not available
                return gevent.ssl.wrap_socket(sock, **self.ssl_options)

            server_hostname = self.ssl_options.get("server_hostname", self._request_host)
            return self.ssl_context.wrap_socket(sock, server_hostname=server_hostname)



"""
Provide HTTPConnection, HTTPSConnection and HTTPResponse implementations ready
to use as drop-in replacements for their counterparts in http.client.
"""

import http.client
from contextlib import contextmanager

import gevent.socket
import gevent.ssl

from geventhttpclient import connectionpool, header, response

_UNKNOWN = getattr(http.client, "_UNKNOWN", "UNKNOWN")


class HTTPLibHeaders(header.Headers):
    def __getitem__(self, key):
        value = super().__getitem__(key)
        if isinstance(value, (list, tuple)):
            return ", ".join(value)
        else:
            return value


class HTTPResponse(response.HTTPSocketResponse):
    def __init__(self, sock, debuglevel=0, method=None, url=None, **kw):
        method = "GET" if method is None else method.upper()
        super().__init__(sock, method=method, **kw)
        self.url = url
        self.chunked = _UNKNOWN
        self.chunk_left = _UNKNOWN

    @property
    def msg(self):
        if hasattr(self, "_msg"):
            return self._msg
        self._msg = HTTPLibHeaders(self._headers_index)
        return self._msg

    @msg.setter
    def msg(self, headers):
        # required by do_open()
        self._msg = headers

    @property
    def fp(self):
        return self

    @property
    def version(self):
        v = self.get_http_version()
        if v == "HTTP/1.1":
            return 11
        return 10

    @property
    def status(self):
        return self.status_code

    @property
    def code(self):
        return self.status_code

    @property
    def reason(self):
        return self.msg

    def _read_status(self):
        return (self.version, self.status_code, self.msg)

    def begin(self):
        pass

    def flush(self):
        self._body_buffer.clear()

    def readable(self):
        return True

    def close(self):
        self.release()

    def isclosed(self):
        return self._sock is None

    def read(self, amt=None):
        return super().read(amt)

    def readinto(self, b):
        raise NotImplementedError()

    def read1(self, n=-1):
        raise NotImplementedError()

    def peek(self, n=-1):
        raise NotImplementedError()

    def fileno(self):
        raise NotImplementedError()

    def getheader(self, name, default=None):
        return self.get(name.lower(), default)

    def getheaders(self):
        return list(self._headers_index.items())

    @property
    def will_close(self):
        return self.message_complete and not self.should_keep_alive()

    def _check_close(self):
        return not self.should_keep_alive()

    # For compatibility with old-style urllib responses. cookielib etc.

    def geturl(self):
        return self.url

    def getcode(self):
        return self.status_code


class HTTPConnection(http.client.HTTPConnection):
    response_class = HTTPResponse
    source_address = None
    _hidden_socket = None

    def connect(self):
        self.sock = gevent.socket.create_connection(
            (self.host, self.port), self.timeout, self.source_address
        )
        if self._tunnel_host:
            self._tunnel()

    def getresponse(self):
        # For recent python versions urllib.request.AbstractHTTPHandler.do_open()
        # insists on closing the socket prematurely, right after receiving a response.
        # So in our case, right after just reading the HTTP headers, the socket gets
        # killed. Therefore, we have two options:
        #
        # 1. We read everything into some buffer, before returning the response object
        # 2. We stop do_open() from messing around with the socket
        #
        # As HTTPSocketResponse is not really intended to buffer big chunks of data,
        # option two remains. Better ideas wellcome.

        if not self.sock and self._hidden_socket is not None:
            self.sock = self._hidden_socket
        resp = super().getresponse()
        self._hidden_socket = self.sock
        self.sock = None
        return resp

    def close(self):
        if not self.sock and self._hidden_socket is not None:
            self.sock = self._hidden_socket
        super().close()


try:
    import gevent.ssl
except ImportError:
    pass
else:

    class HTTPSConnection(HTTPConnection):
        default_port = 443

        def __init__(
            self,
            host,
            port=None,
            key_file=None,
            cert_file=None,
            context=None,
            check_hostname=None,
            **kw,
        ):
            super().__init__(host, port, **kw)
            if key_file is not None or cert_file is not None or check_hostname is not None:
                import warnings

                warnings.warn(
                    "key_file, cert_file and check_hostname are "
                    "deprecated, use a custom context instead.",
                    DeprecationWarning,
                    2,
                )
            self.key_file = key_file
            self.cert_file = cert_file or connectionpool._CA_CERTS
            if context is None:
                context = connectionpool.init_ssl_context(
                    gevent.ssl.create_default_context,
                    self.cert_file,
                    check_hostname=check_hostname,
                )
                # send ALPN extension to indicate HTTP/1.1 protocol
                if self._http_vsn == 11:
                    context.set_alpn_protocols(["http/1.1"])
                # enable PHA for TLS 1.3 connections if available
                if context.post_handshake_auth is not None:
                    context.post_handshake_auth = True
            self._context = context

        def connect(self):
            """Connect to a host on a given (SSL) port."""

            sock = gevent.socket.create_connection(
                (self.host, self.port), self.timeout, self.source_address
            )
            if self._tunnel_host:
                self.sock = sock
                self._tunnel()
            self.sock = gevent.ssl.SSLSocket(
                sock, _context=self._context, server_hostname=self.host
            )


def patch():
    http.client.HTTPConnection = HTTPConnection
    http.client.HTTPResponse = HTTPResponse
    try:
        http.client.HTTPSConnection = HTTPSConnection
    except NameError:
        pass


@contextmanager
def patched():
    """Temporarily patch http.client."""
    http_client_HTTPConnection = http.client.HTTPConnection
    http_client_HTTPResponse = http.client.HTTPResponse
    try:
        http_client_HTTPSConnection = http.client.HTTPSConnection
    except NameError:
        pass
    try:
        http.client.HTTPConnection = HTTPConnection
        http.client.HTTPResponse = HTTPResponse
        try:
            http.client.HTTPSConnection = HTTPSConnection
        except NameError:
            pass
        yield
    finally:
        http.client.HTTPConnection = http_client_HTTPConnection
        http.client.HTTPResponse = http_client_HTTPResponse
        try:
            http.client.HTTPSConnection = http_client_HTTPSConnection
        except NameError:
            pass



# package

__version__ = "2.3.7"  # don't forget to update version in pyproject.toml as well

from geventhttpclient.api import delete, get, head, options, patch, post, put, request
from geventhttpclient.client import HTTPClient
from geventhttpclient.requests import Session
from geventhttpclient.url import URL
from geventhttpclient.useragent import UserAgent
